from __future__ import annotations

import dataclasses
import functools
import re
import typing

import ldap
import ldap.dn


def _escape_backslash(value: str, *, special: str) -> str:
    # escape backslash itself first
    value = value.replace("\\", "\\\\")
    # escape newlines and NULs with special escape sequences
    value = value.replace("\n", "\\n").replace("\r", "\\r").replace("\0", "\\0")
    # escape "specials" by prefixing them with backslash
    pattern_class = re.escape(special)
    return re.sub(f"([{pattern_class}])", r"\\\1", value)


@dataclasses.dataclass(frozen=True)
class DNInfo:
    dn: str
    parts: list[list[tuple[str, str, int]]]  # list of list of (la_attr, la_value, la_flags)

    def __init__(self, *, dn: str) -> None:
        parts = ldap.dn.str2dn(dn, flags=ldap.DN_FORMAT_LDAPV3)
        object.__setattr__(self, "dn", dn)
        object.__setattr__(self, "parts", parts)

    @functools.cached_property
    def domain(self) -> str:
        return ".".join(ava[1] for rdn in self.parts for ava in rdn if ava[0].lower() == "dc")

    def _path(self, *, escape: typing.Callable[[str], str], sep: str) -> str:
        return sep.join(escape(ava[1]) for rdn in reversed(self.parts) for ava in rdn if ava[0].lower() != "dc")

    @functools.cached_property
    def path(self) -> str:
        return self._path(escape=lambda value: _escape_backslash(value, special="/"), sep="/")

    @property
    def full_path(self) -> str:
        domain = self.domain
        path = self.path
        if not path:
            return self.domain
        if not domain:
            return self.path
        return f"{domain}/{path}"
