from __future__ import annotations

import typing

import ldap


class SizeLimitExceeded(Exception):
    pass


Entry = tuple[str, dict[str, list[bytes]]]
Ref = tuple[None, list[str]]
Result = Entry | Ref
Results = list[Result]


def ldap_search_ext(
    ldap_con: ldap.ldapobject.LDAPObject,
    base: str,
    filterstr: str = "(objectClass=*)",
    *,
    scope: int = ldap.SCOPE_SUBTREE,
    attrlist: typing.Optional[typing.Sequence[str]] = None,
    pagelimit: int = 5000,
    sizelimit: int = 0,
    serverctrls: list[ldap.controls.RequestControl] = [],
    **kwargs: typing.Any,
) -> typing.Iterable[Result]:
    """
    Retrieve all results through pagination
    """
    from ldap.controls.libldap import SimplePagedResultsControl

    page_ctrl = SimplePagedResultsControl(criticality=True, size=pagelimit, cookie=b"")
    serverctrls = [page_ctrl] + serverctrls

    def try_get_page() -> tuple[Results, list[ldap.controls.ResponseControl]]:
        response = ldap_con.search_ext(
            base=base,
            scope=scope,
            filterstr=filterstr,
            attrlist=attrlist,
            serverctrls=serverctrls,
            **kwargs,
        )
        _rtype, results, _rmsgid, resp_controls = ldap_con.result3(response)
        # print(f"Ldap search got page: rtype: {_rtype}, results: {len(results)} msgid: {_rmsgid}")
        return results, resp_controls

    def get_page() -> tuple[Results, list[ldap.controls.ResponseControl]]:
        if isinstance(ldap_con, ldap.ldapobject.ReconnectLDAPObject):
            # ReconnectLDAPObject doesn't wrap search_ext / provide search_ext + result3
            return ldap_con._apply_method_s(lambda con: try_get_page())  # type: ignore
        else:
            return try_get_page()

    num_results = 0
    while True:
        if sizelimit:
            # don't get more than 1 result more than we are interested in anyway
            page_ctrl.size = min(pagelimit, sizelimit - num_results + 1)
        results, resp_controls = get_page()

        resp_page_controls = [
            control for control in resp_controls if control.controlType == SimplePagedResultsControl.controlType
        ]
        assert resp_page_controls, "The server ignores RFC 2696 control"

        # forward results from this page
        for result in results:
            if not result[0] is None:
                # don't count refs
                if sizelimit and num_results >= sizelimit:
                    raise SizeLimitExceeded(f"More than {sizelimit} results")
                num_results += 1
            yield result

        # update cookie for next page
        if not resp_page_controls[0].cookie:
            # was last page, done
            break
        page_ctrl.cookie = resp_page_controls[0].cookie
