from __future__ import annotations

import abc
import dataclasses
import getpass
import os
import os.path
import shlex
import subprocess
import typing

import yaml

from ldaptool._utils.dninfo import DNInfo


@dataclasses.dataclass
class Realm:
    # yaml entry key:
    name: str
    # yaml fields:
    domain: str
    servers: list[str]  # space separated in yaml
    forest_root_domain: str  # defaults to domain
    account: typing.Optional[str] = None  # DN or userPrincipalName
    password_file: typing.Optional[str] = None  # password file (default: dervied from account)
    password_folder: typing.Optional[str] = None  # subfolder in password manager

    @staticmethod
    def load(name: str, data: typing.Any) -> Realm:
        assert isinstance(data, dict)
        domain = data.pop("domain")
        servers = data.pop("servers").split()
        forest_root_domain = data.pop("forest_root_domain", domain)
        account = data.pop("account", None)
        password_file = data.pop("password_file", None)
        password_folder = data.pop("password_folder", None)
        return Realm(
            name=name,
            domain=domain,
            servers=servers,
            forest_root_domain=forest_root_domain,
            account=account,
            password_file=password_file,
            password_folder=password_folder,
        )

    def ldap_uri(self, *, gc: bool, tls: bool, server: typing.Optional[str] = None) -> str:
        scheme = "ldaps" if tls else "ldap"
        port = (":3269" if tls else ":3268") if gc else ""  # default ports unless gc
        if not server is None:
            if not server in self.servers:
                raise SystemExit(f"Server {server!r} not listed for realm {self.name}")
            servers = [server]
        else:
            servers = self.servers
        return " ".join(f"{scheme}://{server}.{self.domain}{port}" for server in servers)

    def default_base(self, *, gc: bool) -> str:
        domain = self.forest_root_domain if gc else self.domain
        return ",".join(f"DC={label}" for label in domain.split("."))

    @property
    def password_name(self) -> str:
        """
        Name of password file for the account.

        If password_file wasn't set, it is derived from account:

        If account is using the "email address" format (userPrincipalName),
        the password file is the local part.
        Otherwise it is assumed to be a DN and a full path is extracted from it:
            CN=Bob,OU=SomeDepartment,DC=example,DC=com
        becomes:
            example.com/SomeDepartment/Bob

        If a password_folder was specified, the file is search within it.
        """
        if self.account is None:
            raise ValueError("Require account name to lookup password")
        if not self.password_file is None:
            secretname = self.password_file
        elif "@" in self.account:
            secretname = self.account.split("@", maxsplit=1)[0]
        else:
            secretname = DNInfo(dn=self.account).full_path
        return os.path.join(self.password_folder or "", secretname)


class PasswordManager(abc.ABC):
    @abc.abstractmethod
    def get_password(self, password_name: str) -> str:
        raise NotImplementedError()


@dataclasses.dataclass
class Keyringer(PasswordManager):
    keyring: str
    folder: str

    @staticmethod
    def load(data: typing.Any) -> Keyringer:
        assert isinstance(data, dict)
        keyring = data.pop("keyring")
        folder = data.pop("folder")
        return Keyringer(keyring=keyring, folder=folder)

    def get_password(self, password_name: str) -> str:
        secretname = os.path.join(self.folder, password_name)
        result = subprocess.run(
            [
                "keyringer",
                self.keyring,
                "decrypt",
                secretname,
            ],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            check=True,
            encoding="utf-8",
        )
        return result.stdout.strip()


@dataclasses.dataclass
class Keepass(PasswordManager):
    database: str

    @staticmethod
    def load(data: typing.Any) -> Keepass:
        try:
            import pykeepass  # make sure dependency is present during config loading

            assert pykeepass  # so it isn't unused...
        except ImportError:
            raise Exception(
                "Missing (optional) dependency for keepass support - please install (python package for) pykeepass"
            )
        assert isinstance(
            data, str
        ), "keepass integration expects a single string as configuration with the database location"
        return Keepass(database=data)

    def get_password(self, password_name: str) -> str:
        import pykeepass  # already made sure it is avaiable above

        password = getpass.getpass(f"KeePass password for database {self.database}: ")
        kp = pykeepass.PyKeePass(self.database, password=password)
        entry = kp.find_entries(username=password_name, first=True)
        return entry.password  # type: ignore


@dataclasses.dataclass
class PasswordScript(PasswordManager):
    command: list[str]

    @staticmethod
    def load(data: typing.Any) -> PasswordScript:
        if isinstance(data, str):
            return PasswordScript(command=shlex.split(data))
        elif isinstance(data, list):
            for elem in data:
                assert isinstance(elem, str)
            return PasswordScript(command=data)
        raise ValueError("password-script either takes string or list of strings")

    def get_password(self, password_name: str) -> str:
        result = subprocess.run(
            self.command + [password_name],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            check=True,
            encoding="utf-8",
        )
        return result.stdout.strip()


@dataclasses.dataclass
class Config:
    password_manager: typing.Optional[PasswordManager] = None
    realms: dict[str, Realm] = dataclasses.field(default_factory=dict)

    @staticmethod
    def load() -> Config:
        conf_path = os.path.expanduser("~/.config/ldaptool.yaml")
        if not os.path.exists(conf_path):
            raise SystemExit(f"Missing config file {conf_path}")
        with open(conf_path) as f:
            data = yaml.safe_load(f)
        assert isinstance(data, dict)
        assert "realms" in data
        realms_data = data.pop("realms")
        assert isinstance(realms_data, dict)
        realms = {}
        for name, realm_data in realms_data.items():
            realms[name] = Realm.load(name, realm_data)

        password_manager: typing.Optional[PasswordManager] = None
        if "keyringer" in data:
            if password_manager:
                raise ValueError("Can only set a single password manager")
            password_manager = Keyringer.load(data.pop("keyringer"))
        if "keepass" in data:
            if password_manager:
                raise ValueError("Can only set a single password manager")
            password_manager = Keepass.load(data.pop("keepass"))
        if "password-script" in data:
            if password_manager:
                raise ValueError("Can only set a single password manager")
            password_manager = PasswordScript.load(data.pop("password-script"))

        return Config(realms=realms, password_manager=password_manager)

    def get_password(self, realm: Realm) -> str:
        """
        Return password if password manager is configured.
        Could support other tools as well here.
        """
        if realm.account is None:
            raise RuntimeError("Can't get password without acccount - should use kerberos instead")
        if self.password_manager:
            return self.password_manager.get_password(realm.password_name)

        return getpass.getpass(f"Enter password for {realm.password_name}: ")
